#include <nausparse.h>
#include <vector>
#include <math.h>
#include <fstream>

#ifdef _WIN32
#pragma warning(push)
#pragma warning(disable: 4146 4244 4800)
#include <mpirxx.h>
#pragma warning(pop)
#else
#include <gmpxx.h>
#endif

//convert symmetric matrix a[i,j] to flat symmetric no diagonal upper triangular 
//nxn-matrix by getting the index of [i,j] for i>j
inline int idx_flat(int i, int j)
{
  return int(i*(i-1)/2 + j);
}

//number of elts are the number of upper triangular entries in the adj. mat. without
//the diagonal that is n(n-1)/2. Given this, solve n!
inline int number_of_vertices(const std::vector<bool>& a)
{
  return int((1+sqrt(float(1+8*a.size())))/2);
}

inline size_t number_of_edges(const std::vector<bool>& a)
{
  size_t out = 0;
  for (size_t i=0; i<a.size();++i)
    if (a[i]) out++;
  return out;
}

inline size_t hash_graph(const std::vector<bool>& a)
{
  size_t out=0;
  for (size_t i = 0; i < a.size(); ++i)
    out = (out << 1) + (a[i] ? 1:0);
  return out;
}

inline std::vector<size_t> connected_vertices(const std::vector<bool>& a, const size_t& v, const size_t& n)
{
  //a[i,j] to flat => a[n*i+j] .. i,j runs from 0 to n-1
  //a[i,j] (symmetric) with i<j to lower left half-triangle-no-diagonal flat
  // => a[-i*(1 + i - 2*n)/2 + j - (i+1) ] .. i,j runs from 0 to n-1
  
  //add all in the v-th column
  std::vector<size_t> out;
  for (size_t i=0; i<v;++i)
    if (a[idx_flat(v,i)]) out.push_back(i);
  for (size_t i=v+1; i<n; ++i)
    if (a[idx_flat(i,v)]) out.push_back(i);
  return out;
}

//converts from flat adj mat to nauty sparsegraph
// n is the number of vertices of a
//NOTE: sparseraph out should be freed by SG_FREE(cg);
sparsegraph convert_to_nauty_cg(const std::vector<bool>& a, const size_t& n)
{
  sparsegraph sg;
  SG_INIT(sg);
  size_t no_edges = number_of_edges(a);
  
  SG_ALLOC(sg,n,2*no_edges,"malloc"); //because it is a laman graph!  
  sg.nv = n; //no of vertices
  sg.nde = 2*no_edges;
  
  for (size_t i=0;i<n;++i)
  {
    if (i>0)
      sg.v[i] = sg.v[i-1] + sg.d[i-1];
    else sg.v[i] = 0;
    std::vector<size_t> connected = connected_vertices(a,i,n);
    sg.d[i] = connected.size();
    for (int j=0;j<sg.d[i];++j)
      sg.e[sg.v[i]+j] = connected[j];
  }

  static DEFAULTOPTIONS_SPARSEGRAPH(options);
  
  //debug: options.writeautoms = TRUE;
  options.getcanon = TRUE;

  DYNALLSTAT(int,lab,lab_sz);
  DYNALLSTAT(int,ptn,ptn_sz);
  DYNALLSTAT(int,orbits,orbits_sz);
  DYNALLOC1(int,lab,lab_sz,n,"malloc");
  DYNALLOC1(int,ptn,ptn_sz,n,"malloc");
  DYNALLOC1(int,orbits,orbits_sz,n,"malloc");
  statsblk stats;

  SG_DECL(cg);
  sparsenauty(&sg,lab,ptn,orbits,&options,&stats,&cg);
  
  //freeing DYNALLOC1 objects  
  DYNFREE(orbits,orbits_sz);
  DYNFREE(ptn,ptn_sz);
  DYNFREE(lab,lab_sz);  

  //deallocate lab, ptn, orbits and sg
  SG_FREE(sg);

  return cg;
}

//n= number of vertices
inline void write_to_tgf(const std::vector<bool>& a, const int& n, std::ofstream& f)
{
  using namespace std;
  //first write the vertices
  for (int i=0; i<n; ++i)
    f << i << " " << endl;
  f << "#\n";
  //then write the edges
  size_t idx = 0;
  for (int i=1;i<n; ++i)
  {
    for (int j=0; j<i; ++j)
    {
      if (a[idx]) f << i << " " << j << " " << endl;
      idx++;
    }
  }
}

inline void write_to_tgf(sparsegraph& sg, std::ofstream& f)
{
  using namespace std;
  //first write the vertices
  for (int i=0; i<sg.nv; ++i)
    f << i << " " << endl;
  f << "#\n";
  //int sg.nde = 2*no_edges;
  int k;
  for (int i=0;i<sg.nv-1;++i)
  { 
    for (int j=0;j<sg.d[i];++j)
    {
      k = sg.e[sg.v[i]+j];
      if (k>i) f << i << " " << k << " " << endl;
    }
  }
}

std::vector<bool> canonical_form(const std::vector<bool>& a, int n)
{
  using namespace std;
  sparsegraph cg = convert_to_nauty_cg(a,n);
  //convert cg back to vector of bool
  std::vector<bool> out(a.size(), false);
  
  int k;
  for (int i=0; i<n-1; ++i)
  {
    for (int j=0; j<cg.d[i];++j)
    {
      k= cg.e[cg.v[i]+j];
      if (k>i) out[idx_flat(k,i)]=true;
    }
  }

  SG_FREE(cg); //comment this out if you want to debug.. Brandon says need not free, but this hogs memory if not.
  return out;
}

/*
Henneberg move 1 (only for vector bool format)
n = number of vertices of a
v1 = vertex v1 to do H1 move
v2 = vertex v2 to do H1 move
add an extra node to out
connect the node to v1 (1 edge)
connect the node to v2 (second edge)
return new graph
*/
std::vector<bool> henneberg_1(const std::vector<bool>& a, 
  const size_t& n, const size_t& v1 , const size_t& v2)
{
  std::vector<bool> out(a);
  for (size_t i=0; i<n;++i)
    if ((i==v1)||(i==v2)) out.push_back(true);
    else out.push_back(false);
  
  // in total we have more 1 node and 2 edges
  return out;
}

/*
Henneberg move 2 (only for vector bool format)
Requirement: e_v1 > e_v2, e_v1!=e_v2 and v3 is not equal to e_v1 or e_v2
n = number of vertices of a
e_v1 = vertex of an edge to do H2 move
e_v2 = the other vertex of that edge to do H2 move
v3 = the third vertex not in the edge to do the H2 move

remove edge e_v2,e_v3 but not the nodes associated to them: -1 edge
add an extra node to out
connect the node to e_v1 (1 edge)
connect the node to e_v2 (second edge)
connect the node to v3 (third edge)
*/
std::vector<bool> henneberg_2(const std::vector<bool>& a, const size_t& n,
  size_t e_v1, size_t e_v2, const size_t& v3)
{
  std::vector<bool> out(a);
  if (e_v1 < e_v2) std::swap(e_v2,e_v1);

  out[idx_flat(e_v1,e_v2)] = false;
  
  for (size_t i=0; i<n;++i)
    if ((i==e_v1)||(i==e_v2)||(i==v3)) out.push_back(true);
    else out.push_back(false);
  // in total we have more 1 node and 2 edges
  return out;
}